#!/usr/bin/env python
"""
Test module for PStego.py

"""

__version__ = "$Revision: 1.2 $"
__author__ = "Peter Hendrickson <pdh@wiredyne.com>"

# License:  This code is in the public domain.  Any of this code may be
#           freely reproduced or adapted, even without mentioning the
#           source.

# Warranty: IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR
#           DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
#           DAMAGES ARISING OUT OF THE USE OF THIS CODE, EVEN IF THE
#           AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#           THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES,
#           INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
#           MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
#           CODE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THERE
#           IS NO OBLIGATION WHATSOEVER TO PROVIDE MAINTENANCE,
#           SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

import unittest
import PStego

class PStegoTest(unittest.TestCase):

    def testFactorial(self):
        """Test factorial."""
        self.failUnless(PStego.factorial(1) == 1)
        self.failUnless(PStego.factorial(2) == 2)
        self.failUnless(PStego.factorial(17) == 355687428096000L)
        self.failUnless(PStego.factorial(25) == 15511210043330985984000000L)

    def testPermutedSetSize(self):
        """Test computation of required elements in permutation sets."""
        self.failUnless(PStego.permuted_set_size(1) == 2)
        self.failUnless(PStego.permuted_set_size(2) == 3)
        self.failUnless(PStego.permuted_set_size(100) == 5)
        self.failUnless(PStego.permuted_set_size(1122) == 7)
        self.failUnless(PStego.permuted_set_size(87177805408) == 14)

    def testNatPermConversion(self):
        """Test conversions between permutation numbers and natural numbers."""
        perm_number = PStego.nat_to_perm_number(0, 2)
        zero = PStego.perm_number_to_nat(perm_number)
        self.failUnless(zero == 0)

        perm_number = PStego.nat_to_perm_number(1, 2)
        one = PStego.perm_number_to_nat(perm_number)
        self.failUnless(one == 1)

        for i in xrange(0, 24):
            perm_number = PStego.nat_to_perm_number(i, 3)
            same_i = PStego.perm_number_to_nat(perm_number)
            self.failUnless(i == same_i)

        for i in xrange(0, 120):
            perm_number = PStego.nat_to_perm_number(i, 4)
            same_i = PStego.perm_number_to_nat(perm_number)
            self.failUnless(i == same_i)

        for i in xrange(0, 720):
            perm_number = PStego.nat_to_perm_number(i, 5)
            same_i = PStego.perm_number_to_nat(perm_number)
            self.failUnless(i == same_i)

        for i in xrange(0, 5040):
            perm_number = PStego.nat_to_perm_number(i, 6)
            same_i = PStego.perm_number_to_nat(perm_number)
            self.failUnless(i == same_i)

        for i in xrange(0, 40320):
            perm_number = PStego.nat_to_perm_number(i, 7)
            same_i = PStego.perm_number_to_nat(perm_number)
            self.failUnless(i == same_i)

        for case in [[0, 1, [0]],
                     [1, 1, [1]],
                     [0, 2, [0, 0]],
                     [1, 2, [1, 0]],
                     [2, 2, [0, 1]],
                     [3, 2, [1, 1]],
                     [4, 2, [0, 2]],
                     [5, 2, [1, 2]],
                     [0, 3, [0, 0, 0]],
                     [1, 3, [1, 0, 0]],
                     [3, 3, [1, 1, 0]],
                     [5, 3, [1, 2, 0]],
                     [6, 3, [0, 0, 1]],
                     [12, 3, [0, 0, 2]],
                     [17, 3, [1, 2, 2]],
                     [3301, 7, [1, 0, 2, 2, 3, 4, 0]],
                     [11751, 7, [1, 1, 2, 4, 1, 2, 2]],
                     [22518, 7, [0, 0, 1, 3, 1, 3, 4]],
                     [28545, 7, [1, 1, 1, 4, 3, 4, 5]],
                     [38468, 7, [0, 1, 3, 2, 2, 4, 7]]]:
            (number, cols, perm_number) = case
            result_perm_number = PStego.nat_to_perm_number(number, cols)
            self.failUnless(result_perm_number == perm_number)
            result_int = PStego.perm_number_to_nat(perm_number)
            self.failUnless(result_int == number)

        for i in range(1 + 1):
            for j in range(2 + 1):
                for k in range(3 + 1):
                    for l in range(4 + 1):
                        for m in range(5 + 1):
                            perm_number = [i, j, k, l, m]
                            some_nat = PStego.perm_number_to_nat(perm_number)
                            new_perm_number = \
                              PStego.nat_to_perm_number(some_nat, \
                                                        len(perm_number))
                            self.failUnless(perm_number == new_perm_number)

    def testPermNumberPermConversion(self):
        """Tests conversions between permutation numbers and permutations"""
        for case in [[[0, 0, 1, 2, 5], [4, 3, 0, 1, 2, 5]],
                     [[1, 0, 2, 3, 5], [1, 4, 0, 2, 3, 5]],
                     [[1, 2, 0, 0, 2], [3, 4, 5, 1, 0, 2]],
                     [[0, 1, 3, 1, 2], [4, 0, 3, 5, 1, 2]],
                     [[1, 2, 1, 1, 3], [0, 4, 5, 2, 1, 3]]]:
            (perm_number, permutation) = case
            new_perm_number = PStego.permutation_to_perm_number(permutation)
            new_permutation = PStego.perm_number_to_permutation(perm_number)
            self.failUnless(perm_number == new_perm_number)
            self.failUnless(permutation == new_permutation)
            self.failUnless(len(new_perm_number) == 5)
            self.failUnless(len(permutation) == 6)
            PStego.permutation_check(new_permutation)
            PStego.perm_number_check(new_perm_number)

    def testPermutationNaturalConversion(self):
        """Tests conversions between numbers, permutation and natural."""
        for i in xrange(0, 5040):
            permutation = PStego.nat_to_permutation(i, 7)
            self.failUnless(len(permutation) == 7)
            PStego.permutation_check(permutation)
            new_i = PStego.permutation_to_nat(permutation)
            self.failUnless(i == new_i)

    def testPermutationIntegerConversion(self):
        """Test conversions between permutation numbers and integers."""
        for i in xrange(-2520, 2520):
            permutation = PStego.int_to_permutation(i, 7)
            self.failUnless(len(permutation) == 7)
            PStego.permutation_check(permutation)
            new_i = PStego.permutation_to_int(permutation)
            self.failUnless(i == new_i)

    def testNatStringConversion(self):
        """Test conversions between natural numbers and strings."""

        cases = [ "",
                  "\x00",
                  "\x00\x00",
                  "\x00\x00\x00",
                  "\x00\x00\x00\x00",
                  "\x00\x00\x00\x00\x00",
                  "\xff",
                  "\xff\xff",
                  "\xff\xff\xff",
                  "\xff\xff\xff\xff",
                  "\xff\xff\xff\xff\xff",
                  "\x89\xfc\x4c\x2d\x9b\x3a",
                  "\xc6\x03\xfe\x3a\xee\x47",
                  "\x1a\x06\xee\x77\xbc\xf5",
                  "\xdb\x18\x07\xd6\xeb\xbf",
                  "\x29\xa5\xf0\xb1\xb4\x20",
                  "foobar",
                  "foobar\x00",
                  "foobar\xff",
                  "foobar\x80",
                  "\x00foobar",
                  "\xfffoobar",
                  "\x80foobar",
                  "William Shakespeare",
                  "Now is the time for all good men to come to " + \
                  "the aid of their party." ]

        for case in cases:
            nat = PStego.string_to_nat(case)
            copy = PStego.nat_to_string(nat)
            if case != copy:
                print "Case:", case
            self.failUnless(case == copy)

unittest.main()
